import asyncio
import json
import random
from itertools import cycle, product
from math import ceil, gcd, lcm
from pathlib import Path
from typing import Any

import json_repair
import yaml
from openai import AsyncOpenAI
from openai.types.chat import ChatCompletion
from openai.types.images_response import ImagesResponse
from pydantic import BaseModel, Field, PrivateAttr, ValidationError

from deeppresenter.utils.constants import (
    CONTEXT_LENGTH_LIMIT,
    MCP_CALL_TIMEOUT,
    PACKAGE_DIR,
    PIXEL_MULTIPLE,
    RETRY_TIMES,
)
from deeppresenter.utils.log import debug, info, logging_openai_exceptions


def get_json_from_response(response: str) -> dict | list:
    """
    Extract JSON from a text response.

    Args:
        response (str): The response text.

    Returns:
        Dict|List: The extracted JSON.

    Raises:
        Exception: If JSON cannot be extracted from the response.
    """

    assert isinstance(response, str) and len(response) > 0, (
        "response must be a non-empty string"
    )
    response = response.strip()
    try:
        return json.loads(response)
    except Exception:
        pass

    # Try to find JSON by looking for matching braces
    open_braces = []
    close_braces = []

    for i, char in enumerate(response):
        if char == "{" or char == "[":
            open_braces.append(i)
        elif char == "}" or char == "]":
            close_braces.append(i)

    for i, j in product(open_braces, reversed(close_braces)):
        if i > j:
            continue
        try:
            json_obj = json.loads(response[i : j + 1])
            if isinstance(json_obj, (dict, list)):
                return max(
                    json_obj, json_repair.loads(response), key=lambda x: len(str(x))
                )
        except Exception:
            pass

    return json_repair.loads(response)


class Endpoint(BaseModel):
    """LLM Endpoint Configuration"""

    base_url: str = Field(description="API base URL")
    model: str = Field(description="Model name")
    api_key: str = Field(description="API key")
    client_kwargs: dict[str, Any] = Field(
        default_factory=dict, description="Client parameters"
    )
    sampling_parameters: dict[str, Any] = Field(
        default_factory=dict, description="Sampling parameters"
    )
    _client: AsyncOpenAI = PrivateAttr()

    def model_post_init(self, _) -> None:
        self._client = AsyncOpenAI(
            api_key=self.api_key,
            base_url=self.base_url,
            **self.client_kwargs,
        )

    async def call(
        self,
        messages: list[dict[str, Any]],
        soft_response_parsing: bool,
        response_format: type[BaseModel] | None = None,
        tools: list[dict[str, Any]] | None = None,
    ) -> ChatCompletion:
        """Execute a chat or tool call using the endpoint client"""
        if tools is not None:
            response = await self._client.chat.completions.create(
                model=self.model,
                messages=messages,
                tools=tools,
                tool_choice="auto",
                **self.sampling_parameters,
            )
        elif not soft_response_parsing and response_format is not None:
            response: ChatCompletion = await self._client.chat.completions.parse(
                model=self.model,
                messages=messages,
                response_format=response_format,
                **self.sampling_parameters,
            )
        else:
            response: ChatCompletion = await self._client.chat.completions.create(
                model=self.model,
                messages=messages,
                **self.sampling_parameters,
            )
        assert response.choices is not None and len(response.choices) > 0, (
            f"No choices returned from the model, got {response}"
        )
        message = response.choices[0].message
        debug(f"Response from {self.model}: {message}")
        if response_format is not None:
            message.content = response_format(
                **get_json_from_response(message.content)
            ).model_dump_json(indent=2)
        assert tools is None or len(message.tool_calls or []), (
            f"No tool call returned from the model, got {message}"
        )
        assert message.tool_calls or message.content, (
            "Empty content returned from the model"
        )
        return response


class LLM(BaseModel):
    """LLM Client Manager"""

    base_url: str | None = Field(default=None, description="API base URL")
    model: str | None = Field(default=None, description="Model name")
    api_key: str | None = Field(default=None, description="API key")
    identifier: str | None = Field(
        default=None,
        description="Optional identifier for the model instance, this will override property `model_name`",
    )
    is_multimodal: bool | None = Field(
        default=None, description="Whether the model is multimodal"
    )
    max_concurrent: int | None = Field(
        default=None, description="Maximum concurrency limit"
    )
    client_kwargs: dict[str, Any] = Field(
        default_factory=dict, description="Client parameters"
    )
    sampling_parameters: dict[str, Any] = Field(
        default_factory=dict, description="Sampling parameters"
    )
    endpoints: list[dict[str, Any]] = Field(
        default_factory=list,
        description="Additional endpoints for alternating retries",
    )
    soft_response_parsing: bool = Field(
        default=False,
        description="Enable soft parsing: parse response content as JSON directly instead of using completion.parse",
    )
    min_image_size: int | None = Field(
        default=None,
        description="Minimum image size (width * height) for generation, smaller images will be resized proportionally",
    )
    secret_logging: bool = Field(
        default=False, description="Logging detailed endpoint (API key included)"
    )

    _semaphore: asyncio.Semaphore = PrivateAttr()
    _endpoints: list[Endpoint] = PrivateAttr(default_factory=list)

    model_config = {"arbitrary_types_allowed": True}

    @property
    def model_name(self) -> str:
        return self.identifier or self._endpoints[0].model.split("/")[-1].split(":")[0]

    def model_post_init(self, context) -> None:
        """Initialize semaphore and endpoints"""
        self._semaphore = asyncio.Semaphore(self.max_concurrent or 10000)
        if self.model:
            self._endpoints.insert(
                0,
                Endpoint(
                    base_url=self.base_url,
                    model=self.model,
                    api_key=self.api_key,
                    client_kwargs=self.client_kwargs,
                    sampling_parameters=self.sampling_parameters,
                ),
            )
        for endpoint in self.endpoints:
            self._endpoints.append(Endpoint(**endpoint))
        assert len(self._endpoints) >= 1, "At least one endpoint must be configured"

        model_lower = self._endpoints[0].model.lower()
        if self.is_multimodal is None:
            if any(word in model_lower for word in ("gpt", "claude", "gemini", "vl")):
                self.is_multimodal = True
                debug(
                    f"Model {self._endpoints[0].model} is detected as multimodal model, setting `is_multimodal` to True"
                )
            else:
                self.is_multimodal = False

        return super().model_post_init(context)

    async def run(
        self,
        messages: list[dict[str, Any]] | str,
        response_format: type[BaseModel] | None = None,
        tools: list[dict[str, Any]] | None = None,
        retry_times: int = RETRY_TIMES,
    ) -> ChatCompletion:
        """Unified interface for chat and tool calls with alternating retry"""
        if isinstance(messages, str):
            messages = [{"role": "user", "content": messages}]

        errors = []
        iter_endpoints = cycle(self._endpoints)
        async with self._semaphore:
            for _ in range(retry_times):
                endpoint = next(iter_endpoints)
                try:
                    return await endpoint.call(
                        messages,
                        self.soft_response_parsing,
                        response_format,
                        tools,
                    )
                except (AssertionError, ValidationError) as e:
                    errors.append(f"[{endpoint.model}] {e}")
                except Exception as e:
                    errors.append(f"[{endpoint.model}] {e}")
                    if self.secret_logging:
                        identifider = endpoint
                    else:
                        identifider = endpoint.model
                    logging_openai_exceptions(identifider, e)
        raise ValueError(f"All models failed after {retry_times} retries:\n{errors}")

    async def generate_image(
        self,
        prompt: str,
        width: int,
        height: int,
        retry_times: int = RETRY_TIMES,
        pixel_multiple: int = PIXEL_MULTIPLE,
    ) -> ImagesResponse:
        """Unified interface for image generation"""
        if self.min_image_size is not None and (width * height) < int(
            self.min_image_size
        ):
            ratio = (int(self.min_image_size) / (width * height)) ** 0.5
            width = int(width * ratio)
            height = int(height * ratio)
        assert (width % PIXEL_MULTIPLE == 0) and (height % PIXEL_MULTIPLE == 0), (
            f"Image width and height must be a multiple of {pixel_multiple}"
        )
        async with self._semaphore:
            errors = []
            random.shuffle(self._endpoints)
            for retry_idx in range(retry_times):
                # t2i is stateless
                endpoint = self._endpoints[retry_idx % len(self._endpoints)]
                try:
                    response = await endpoint._client.images.generate(
                        prompt=prompt,
                        model=endpoint.model,
                        size=f"{width}x{height}",
                        timeout=MCP_CALL_TIMEOUT // 5,
                        **endpoint.sampling_parameters,
                    )
                    assert len(response.data) >= 1, (
                        f"Expected at least an image response, got {response}"
                    )
                    return response

                except (AssertionError, ValidationError) as e:
                    errors.append(f"[{endpoint.model}] {e}")
                except Exception as e:
                    errors.append(f"[{endpoint.model}] {e}")
                    if self.secret_logging:
                        identifider = endpoint
                    else:
                        identifider = endpoint.model
                    logging_openai_exceptions(identifider, e)
            raise ValueError(f"All models failed after {retry_times} retries: {errors}")

    async def validate(self):
        endpoint = self._endpoints[0]
        models = await endpoint._client.models.list()
        # ? This for compatibility with google generative ai
        if not any(model.id.endswith(endpoint.model) for model in models.data):
            raise Exception(
                f"Model {endpoint.model} is not available at {endpoint.base_url}, please check your apikey or {PACKAGE_DIR / 'config.yaml'}\n"
            )


class DeepPresenterConfig(BaseModel):
    """DeepPresenter Global Configuration"""

    # config
    offline_mode: bool = Field(
        default=False, description="Enable offline mode, disable all network requests"
    )
    file_path: str = Field(description="Configuration file path")
    mcp_config_file: str = Field(
        description="MCP configuration file", default=PACKAGE_DIR / "mcp.json"
    )
    context_folding: bool = Field(
        default=True, description="Enable context management and auto summarization"
    )
    context_window: int | None = Field(
        default=None,
        description="Context window for context management, if not set, use the default value",
    )
    max_context_folds: int = Field(
        default=5, description="Maximum number of folds for context management"
    )
    heavy_reflect: bool = Field(
        default=False,
        description="Enable heavy reflection, use rendered slide image for reflective design",
    )

    # llms
    research_agent: LLM = Field(description="Research agent model configuration")
    design_agent: LLM = Field(description="Design agent model configuration")
    long_context_model: LLM = Field(description="Long context model configuration")
    vision_model: LLM | None = Field(
        default=None, description="Vision model configuration"
    )
    t2i_model: LLM | None = Field(
        default=None, description="Text-to-image model configuration"
    )

    def model_post_init(self, context):
        if self.context_window is None:
            if self.context_folding:
                self.context_window = CONTEXT_LENGTH_LIMIT // self.max_context_folds
            else:
                self.context_window = CONTEXT_LENGTH_LIMIT

        if self.context_folding:
            debug(
                f"Context folding is enabled, context window: {self.context_window}, max folds: {self.max_context_folds}"
            )
        else:
            debug(f"Context folding is disabled, context window: {self.context_window}")

        return super().model_post_init(context)

    @classmethod
    def load_from_file(cls, config_path: str | None = None) -> "DeepPresenterConfig":
        """Load configuration from file"""
        if config_path:
            config_file = Path(config_path)
        else:
            config_file = PACKAGE_DIR / "config.yaml"

        if not config_file.exists():
            raise FileNotFoundError(f"Configuration file {config_file} does not exist")
        config_data = {}
        with open(config_file, encoding="utf-8") as f:
            config_data = yaml.safe_load(f) or {}

        config_data["file_path"] = str(config_file.resolve())
        return cls(**config_data)

    async def validate_llms(self):
        # ? t2i endpoints might not support this api
        tasks = [
            self.research_agent.validate(),
            self.design_agent.validate(),
            self.long_context_model.validate(),
        ]
        if self.vision_model is not None:
            tasks.append(self.vision_model.validate())
        await asyncio.gather(*tasks)

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)



import asyncio
import os
import tempfile
import zipfile
from pathlib import Path

import aiohttp


async def parse_pdf_offline(pdf_path: str, output_path: str, url: str) -> None:
    """Parse PDF using a local/compatible MinerU endpoint."""
    os.makedirs(output_path, exist_ok=True)
    pdf_path_obj = Path(pdf_path)

    async with aiohttp.ClientSession() as session:
        form = aiohttp.FormData()
        form.add_field(
            "pdf",
            pdf_path_obj.read_bytes(),
            filename=pdf_path_obj.name,
            content_type="application/pdf",
        )

        async with session.post(url, data=form) as resp:
            if resp.status != 200:
                await _raise_parsedoc_error(resp)
            content = await resp.read()

    _extract_zip_bytes(content, output_path)


async def parse_pdf_online(
    pdf_path: str, output_path: str, token: str, model_version: str = "vlm"
) -> None:
    """Parse PDF using MinerU external API

    Args:
        pdf_path: PDF file path
        output_path: Output directory
        token: API Token
        model_version: Model version (vlm/pipeline)
    """
    os.makedirs(output_path, exist_ok=True)
    pdf_path = Path(pdf_path)

    async with aiohttp.ClientSession() as session:
        batch_id, upload_url, upload_headers = await _request_upload_url(
            session, pdf_path.name, pdf_path.stem[:128], model_version, token
        )

        await _upload_file(session, upload_url, pdf_path, upload_headers)

        zip_url = await _poll_result(session, batch_id, token)

        await _download_and_extract(session, zip_url, output_path)


async def _request_upload_url(
    session: aiohttp.ClientSession,
    filename: str,
    data_id: str,
    model_version: str,
    token: str,
) -> tuple[str, str, dict[str, str] | None]:
    """Request upload URL, returns (batch_id, upload_url, upload_headers)"""
    headers = {"Content-Type": "application/json", "Authorization": f"Bearer {token}"}
    payload = {
        "files": [{"name": filename, "data_id": data_id}],
        "model_version": model_version,
    }

    async with session.post(
        "https://mineru.net/api/v4/file-urls/batch", headers=headers, json=payload
    ) as resp:
        resp.raise_for_status()
        result = await resp.json()
        if result["code"] != 0:
            raise RuntimeError(
                f"Failed to request upload URL: {result.get('msg', 'Unknown error')}"
            )

        data = result["data"]
        upload_headers = data.get("headers", [None])[0] if "headers" in data else None
        return data["batch_id"], data["file_urls"][0], upload_headers


async def _upload_file(
    session: aiohttp.ClientSession,
    upload_url: str,
    pdf_path: Path,
    headers: dict[str, str] | None = None,
) -> None:
    """Upload PDF file to OSS"""
    file_data = pdf_path.read_bytes()

    upload_headers = headers if headers else {}

    async with session.put(
        upload_url,
        data=file_data,
        headers=upload_headers,
        skip_auto_headers={"Content-Type"},
    ) as resp:
        resp.raise_for_status()


async def _poll_result(
    session: aiohttp.ClientSession, batch_id: str, token: str
) -> str:
    """Poll parsing result, returns download URL"""
    headers = {"Authorization": f"Bearer {token}"}
    url = f"https://mineru.net/api/v4/extract-results/batch/{batch_id}"

    while True:
        async with session.get(url, headers=headers) as resp:
            resp.raise_for_status()
            result = await resp.json()

            if result["code"] != 0:
                raise RuntimeError(
                    f"Query failed: {result.get('msg', 'Unknown error')}"
                )

            extract = result["data"]["extract_result"][0]
            state = extract["state"]

            if state == "done":
                return extract["full_zip_url"]
            elif state == "failed":
                raise RuntimeError(
                    f"Parsing failed: {extract.get('err_msg', 'Unknown error')}"
                )

            await asyncio.sleep(5)


async def _download_and_extract(
    session: aiohttp.ClientSession, zip_url: str, output_path: str
) -> None:
    """Download and extract result"""
    async with session.get(zip_url) as resp:
        resp.raise_for_status()
        content = await resp.read()

    _extract_zip_bytes(content, output_path)


async def _raise_parsedoc_error(resp: aiohttp.ClientResponse) -> None:
    """Raise a RuntimeError with parsed error content."""
    try:
        payload = await resp.json()
    except Exception:
        payload = await resp.text()
    raise RuntimeError(payload)


def _extract_zip_bytes(content: bytes, output_path: str) -> None:
    """Extract zip bytes into output_path."""
    with tempfile.NamedTemporaryFile(delete=False, suffix=".zip") as tmp:
        tmp.write(content)
        zip_path = tmp.name

    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        all_names = [name for name in zip_ref.namelist() if name.strip()]
        top_level = {name.split("/", 1)[0] for name in all_names}

        if len(top_level) == 1 and all("/" in name for name in all_names):
            prefix = list(top_level)[0] + "/"
        else:
            prefix = ""

        for member in zip_ref.infolist():
            if not member.is_dir():
                rel_path = (
                    member.filename.removeprefix(prefix) if prefix else member.filename
                )
                dest_path = os.path.join(output_path, rel_path)
                os.makedirs(os.path.dirname(dest_path), exist_ok=True)
                with zip_ref.open(member) as src, open(dest_path, "wb") as dst:
                    dst.write(src.read())

    os.unlink(zip_path)



import asyncio
import json
import uuid
from abc import abstractmethod
from collections.abc import AsyncGenerator
from datetime import datetime
from pathlib import Path
from typing import Literal

import jsonlines
import yaml
from jinja2 import Template
from jinja2.runtime import StrictUndefined
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageFunctionToolCall as ToolCall,
)
from pydantic import BaseModel

from deeppresenter.agents.env import AgentEnv
from deeppresenter.utils.config import (
    LLM,
    DeepPresenterConfig,
    get_json_from_response,
)
from deeppresenter.utils.constants import (
    AGENT_PROMPT,
    CONTEXT_MODE_PROMPT,
    CONTINUE_MSG,
    HALF_BUDGET_NOTICE_MSG,
    HIST_LOST_MSG,
    LAST_ITER_MSG,
    MAX_LOGGING_LENGTH,
    MAX_TOOLCALL_PER_TURN,
    MEMORY_COMPACT_MSG,
    OFFLINE_PROMPT,
    PACKAGE_DIR,
    URGENT_BUDGET_NOTICE_MSG,
)
from deeppresenter.utils.log import (
    debug,
    info,
    timer,
)
from deeppresenter.utils.typings import (
    ChatMessage,
    Cost,
    InputRequest,
    Role,
    RoleConfig,
)


class Agent:
    def __init__(
        self,
        config: DeepPresenterConfig,
        agent_env: AgentEnv,
        workspace: Path,
        language: Literal["zh", "en"],
        config_file: str | None = None,
        keep_reasoning: bool = True,
    ):
        self.name = self.__class__.__name__
        self.cost = Cost()
        self.context_length = 0
        self.context_warning = 0
        self.workspace = workspace
        self.agent_env = agent_env
        self.language = language
        self.keep_reasoning = keep_reasoning
        self.context_window = config.context_window
        self.max_context_turns = config.max_context_folds
        config_file = (
            Path(config_file)
            if config_file
            else PACKAGE_DIR / "roles" / f"{self.name}.yaml"
        )
        if not config_file.exists():
            raise FileNotFoundError(f"Cannot found role config file at: {config_file} ")

        # Setting basic context
        with open(config_file, encoding="utf-8") as f:
            config_data = yaml.safe_load(f)
        self.role_config = RoleConfig(**config_data)
        self.llm: LLM = config[self.role_config.use_model]
        self.model = self.llm.model_name
        self._setup_toolset()
        if language not in self.role_config.system:
            raise ValueError(f"Language '{language}' not found in system prompts")
        self.error_history: list[ToolCall | ChatMessage] = []
        self.research_iter = 0
        if config.context_folding:
            self.context_warning = -1

        # Setting tools and interative context
        self.system = self.role_config.system[language]
        self.prompt: Template = Template(
            self.role_config.instruction, undefined=StrictUndefined
        )
        # ? for those agents equipped with sandbox only
        if any(t["function"]["name"] == "execute_command" for t in self.tools):
            self.system += AGENT_PROMPT.format(
                workspace=self.workspace,
                cutoff_len=self.agent_env.cutoff_len,
                time=datetime.now().strftime("%Y-%m-%d"),
                max_toolcall_per_turn=MAX_TOOLCALL_PER_TURN,
            )

        if config.offline_mode:
            self.system += OFFLINE_PROMPT

        if config.context_folding:
            self.system += CONTEXT_MODE_PROMPT

        self.chat_history: list[ChatMessage] = [
            ChatMessage(role=Role.SYSTEM, content=self.system)
        ]
        available_tools = [tool["function"]["name"] for tool in self.tools]
        debug(
            f"{self.name} Agent got {len(self.tools)} tools: {', '.join(available_tools)}"
        )

    def _setup_toolset(self):
        toolset = self.role_config.toolset
        if toolset.include_tool_servers == "all":
            toolset.include_tool_servers = list(self.agent_env._server_tools)
        for server in toolset.include_tool_servers:
            assert server in self.agent_env._server_tools, (
                f"Server {server} is not available"
            )
        for tool in toolset.include_tools + toolset.exclude_tools:
            assert tool in self.agent_env._tools_dict, f"Tool {tool} is not available"
        self.tools = []
        for server in toolset.include_tool_servers:
            if server not in toolset.exclude_tool_servers:
                for tool in self.agent_env._server_tools[server]:
                    if tool not in toolset.exclude_tools:
                        self.tools.append(self.agent_env._tools_dict[tool])

        for tool_name, tool in self.agent_env._tools_dict.items():
            if tool_name in toolset.include_tools:
                self.tools.append(tool)

    async def chat(
        self,
        message: ChatMessage,
        response_format: type[BaseModel] | None = None,
        **chat_kwargs,
    ) -> ChatMessage:
        if len(self.chat_history) == 1:
            self.chat_history.append(
                ChatMessage(role=Role.USER, content=self.prompt.render(**chat_kwargs))
            )
            self.log_message(self.chat_history[-1])
        self.chat_history.append(message)
        self.log_message(self.chat_history[-1])
        with timer(f"{self.name} Agent LLM chat"):
            response = await self.llm.run(
                messages=self.chat_history,
                response_format=response_format,
            )
            if response.usage is not None:
                self.cost += response.usage
                self.context_length = response.usage.total_tokens
            self.chat_history.append(
                ChatMessage(
                    role=Role.ASSISTANT,
                    content=response.choices[0].message.content,
                    cost=response.usage,
                    reasoning=getattr(response.choices[0].message, "reasoning", None)
                    if self.keep_reasoning
                    else None,
                )
            )
            self.log_message(self.chat_history[-1])
            return self.chat_history[-1]

    async def action(
        self,
        **chat_kwargs,
    ):
        """Tool calling interface"""

        if len(self.chat_history) == 1:
            self.chat_history.append(
                ChatMessage(
                    role=Role.USER,
                    content=self.prompt.render(**chat_kwargs),
                )
            )
            self.log_message(self.chat_history[-1])

        with timer(f"{self.name} Agent LLM call"):
            response = await self.llm.run(
                messages=self.chat_history,
                tools=self.tools,
            )
            if response.usage is not None:
                self.cost += response.usage
                self.context_length = response.usage.total_tokens
            agent_message: ChatCompletionMessage = response.choices[0].message
        self.chat_history.append(
            ChatMessage(
                role=Role.ASSISTANT,
                content=agent_message.content,
                cost=response.usage,
                tool_calls=agent_message.tool_calls,
                reasoning=getattr(agent_message, "reasoning", None)
                if self.keep_reasoning
                else None,
            )
        )
        self.log_message(self.chat_history[-1])
        return self.chat_history[-1]

    @abstractmethod
    async def loop(
        self, req: InputRequest, *args, **kwargs
    ) -> AsyncGenerator[str | ChatMessage, None]:
        """
        Loop interface, return the message or the outcome filepath of the agent.
        """

    @abstractmethod
    async def finish(self, result: str):
        """This function defines when and how should an agent finish their tasks, combined with outcome check"""

    async def execute(self, tool_calls: list[ToolCall]) -> str | list[ChatMessage]:
        coros = []
        observations: list[ChatMessage] = []
        used_tools = set()
        finish_id = None
        outcome = None
        for t in tool_calls:
            arguments = t.function.arguments
            if len(arguments) == 0:
                arguments = None
            else:
                try:
                    assert len(tool_calls) <= MAX_TOOLCALL_PER_TURN, (
                        f"Too many tool calls ({len(tool_calls)}), max allowed is {MAX_TOOLCALL_PER_TURN}"
                    )
                    arguments = get_json_from_response(t.function.arguments)
                    if t.function.name == "finalize":
                        arguments["agent_name"] = self.name
                        finish_id = t.id
                        assert "outcome" in arguments, (
                            "Finalize tool call must have an outcome"
                        )
                        outcome = arguments["outcome"]
                    assert isinstance(arguments, dict), (
                        f"Tool call arguments must be a dict or empty, while {arguments} is given"
                    )
                    t.function.arguments = json.dumps(arguments, ensure_ascii=False)
                except AssertionError as e:
                    observations.append(
                        ChatMessage(
                            role=Role.TOOL,
                            content=str(e),
                            tool_call_id=t.id,
                            is_error=True,
                        )
                    )
                    info(f"Tool call `{t.function}` encountered error: {e}")
                    continue
            used_tools.add(t.function.name)
            info(f"{self.name} Agent calling tool `{t.function.name}`")
            coros.append(self.agent_env.tool_execute(t))

        observations.extend(await asyncio.gather(*coros))
        for obs in observations:
            if obs.has_image:
                if "gemini" in self.model.lower() or "qwen" in self.model.lower():
                    obs.role = Role.USER
                if "claude" in self.model.lower():
                    oai_b64 = obs.content[0]["image_url"]["url"]
                    obs.content = [
                        {
                            "type": "image",
                            "source": {
                                "type": "base64",
                                "media_type": oai_b64.split(";")[0].split(":")[1],
                                "data": oai_b64.split(",")[1],
                            },
                        }
                    ]

        self.chat_history.extend(observations)

        for t, o in zip(
            sorted(tool_calls, key=lambda x: x.id),
            sorted(observations, key=lambda x: x.id),
        ):
            if o.is_error:
                self.error_history.append(t)
                self.error_history.append(o)

        if finish_id is not None:
            for obs in observations:
                if obs.tool_call_id == finish_id and obs.text == outcome:
                    info(f"{self.name} Agent finished with result: {obs.text}")
                    return obs.text

        if (
            self.context_warning == 0
            and self.context_length > self.context_window * 0.5
        ):
            self.context_warning += 1
            observations[0].content.insert(0, HALF_BUDGET_NOTICE_MSG)
        elif (
            self.context_warning == 1
            and self.context_length > self.context_window * 0.8
        ):
            observations[0].content.insert(0, URGENT_BUDGET_NOTICE_MSG)
            self.context_warning = 2

        for obs in observations:
            self.log_message(obs)

        if self.context_length > self.context_window:
            if self.context_warning == -1:
                await self.compact_history()
            else:
                raise RuntimeError(
                    f"{self.name} agent exceeded context window: {self.context_length}/{self.context_window}"
                )
        return observations

    def log_message(self, msg: ChatMessage):
        if len(msg.text) < MAX_LOGGING_LENGTH:
            debug(f"{self.name}: {msg.text}")
        else:
            debug(f"{self.name}: {msg.text[:MAX_LOGGING_LENGTH]}...")

    async def compact_history(self, keep_head: int = 10, keep_tail: int = 4):
        """Summarize the history."""
        # ? it's 10 = system + user + (thinking, read, design, write)*2
        if keep_head + keep_tail > len(self.chat_history):
            return

        if self.research_iter == self.max_context_turns:
            return

        self.save_history(message_only=True)
        self.research_iter += 1
        head, tail = self._split_history(keep_head, keep_tail)
        summary_ask = ChatMessage(
            role=Role.USER, content=MEMORY_COMPACT_MSG.format(language=self.language)
        )
        response = await self.llm.run(
            self.chat_history + [summary_ask],
            tools=self.tools,
        )
        agent_message = response.choices[0].message
        summary_message = ChatMessage(
            id=f"context_fold_{uuid.uuid4().hex[:8]}",
            role=agent_message.role,
            content=agent_message.content,
            tool_calls=agent_message.tool_calls,
            reasoning=getattr(agent_message, "reasoning", None)
            if self.keep_reasoning
            else None,
        )
        debug(
            f"Summary of Resarch Iter {self.research_iter:02d}: \n"
            + summary_message.text
        )
        tasks = [
            self.agent_env.tool_execute(tc) for tc in summary_message.tool_calls or []
        ]
        observations = await asyncio.gather(*tasks)
        observations[-1].content.append(CONTINUE_MSG)
        if self.research_iter == self.max_context_turns:
            observations[-1].content.append(LAST_ITER_MSG)
        new_tail = [
            summary_ask,
            summary_message,
            *observations,
        ]
        self.chat_history = head + tail + new_tail

    def _split_history(self, keep_head, keep_tail):
        # ensure the left context window contains the paired tool call and tool call result
        head = []
        for msg in self.chat_history:
            if len(head) < keep_head or msg.role == Role.TOOL:
                head.append(msg)
            else:
                break
        head[-1].content.append(HIST_LOST_MSG)

        tail = self.chat_history[-keep_tail:]
        for i, m in enumerate(tail):
            if m.role == Role.ASSISTANT and m not in head:
                tail = tail[i:]
                break
        else:
            tail = []

        return head, tail

    def save_history(self, hist_dir: Path | None = None, message_only: bool = False):
        hist_dir = hist_dir or self.workspace / ".history"
        hist_dir.mkdir(parents=True, exist_ok=True)

        history_file = hist_dir / f"{self.name}-history.jsonl"
        if self.research_iter >= 0:
            history_file = (
                hist_dir / f"{self.name}-{self.research_iter:02d}-history.jsonl"
            )
        with jsonlines.open(history_file, mode="w") as writer:
            for message in self.chat_history:
                writer.write(message.model_dump())

        if message_only:
            return

        config_file = hist_dir / f"{self.name}-config.json"
        with open(config_file, "w", encoding="utf-8") as f:
            json.dump(
                {
                    "name": self.name,
                    "model": self.model,
                    "context_window": self.context_length,
                    "cost": self.cost.model_dump(),
                    "tools": self.tools,
                },
                f,
                ensure_ascii=False,
                indent=2,
            )

        if self.error_history:
            error_file = hist_dir / f"{self.name}-errors.jsonl"
            with jsonlines.open(error_file, mode="w") as writer:
                for msg in self.error_history:
                    writer.write(msg.model_dump())

        debug(
            f"{self.name} done | cost:{self.cost} ctx:{self.context_length} | history:{history_file.name} config:{config_file.name}"
        )



"""DeepPresenter - An agentic and reflective presentation generation system"""

__version__ = "1.1.16"
__author__ = "Hao Zheng"
__email__ = "wszh712811@gmail.com"

import os

assert os.name == "posix", "DeepPresenter only supports Linux and macOS"



import base64
import os
import re
import sys
import uuid
from pathlib import Path

from fastmcp import FastMCP
from markitdown import MarkItDown
from PIL import Image

from deeppresenter.utils.log import set_logger, warning
from deeppresenter.utils.mineru_api import parse_pdf_offline, parse_pdf_online

mcp = FastMCP(name="Any2Markdown")

IMAGE_EXTENSIONS = [
    "bmp",
    "jpg",
    "jpeg",
    "pgm",
    "png",
    "ppm",
    "tif",
    "tiff",
    "webp",
]
MINERU_API_URL = os.getenv("MINERU_API_URL", None)
MINERU_API_KEY = os.getenv("MINERU_API_KEY", None)


@mcp.tool()
async def convert_to_markdown(file_path: str, output_folder: str) -> dict:
    """Convert a file to markdown, it could accept pdf, docx, doc, etc.
    Args:
        file_path: The path of the file to be converted
        output_folder: The folder to save the converted markdown and images, should be empty or not exist

    Returns:
        The converted results, with file saved to the specified path
    """

    output_path = Path(output_folder)
    output_path.mkdir(parents=True, exist_ok=True)
    assert len(list(output_path.iterdir())) == 0, (
        f"Output folder {output_folder} is not empty"
    )
    assert os.path.exists(file_path), f"Error: file {file_path} does not exist"

    markdown_file = output_path / f"{Path(file_path).stem}.md"

    if file_path.lower().endswith(".pdf") and (MINERU_API_KEY or MINERU_API_URL):
        if MINERU_API_KEY:
            await parse_pdf_online(file_path, str(output_path), MINERU_API_KEY)
        elif MINERU_API_URL:
            await parse_pdf_offline(file_path, str(output_path), MINERU_API_URL)
        for f in output_path.glob("*"):
            if f.name.lower().endswith(".md"):
                os.rename(f, str(markdown_file))
            elif f.name.lower().endswith((".json", ".pdf")):
                os.remove(f)
        with open(str(markdown_file), encoding="utf-8") as f:
            markdown = f.read()
    else:
        conver_result = MarkItDown().convert_local(file_path, keep_data_uris=True)
        markdown = parse_base64_images(
            conver_result.text_content, output_path / "images"
        )

    for match in re.findall(r"!\[.*?\]\((.*?)\)", markdown):
        local_path = match.split()[0].strip("\"'")
        p = Path(local_path)
        if (output_path / local_path).exists():
            p = output_path / local_path
        if p.exists():
            markdown = markdown.replace(local_path, str(p.resolve()))
    with open(str(markdown_file), "w", encoding="utf-8") as f:
        f.write(markdown)

    images = output_path.glob("images/*")
    images_with_info = []
    for img_path in images:
        try:
            with Image.open(img_path) as img:
                images_with_info.append((img_path, *img.size))
        except Exception:
            continue

    images_with_info.sort(key=lambda x: int(x[1]), reverse=True)

    return {
        "success": True,
        "markdown_file": str(markdown_file),
        "images": f"Found {len(images_with_info)} images\n"
        + "".join([f"- {img[0]}: {img[1]}x{img[2]}\n" for img in images_with_info]),
    }


def parse_base64_images(markdown: str, image_dir: Path) -> str:
    """Save base64 images to local, and convert those links to local paths"""
    image_dir.mkdir(exist_ok=True, parents=True)
    for image_match in re.finditer(
        r"!\[([^\]]*)\]\((data:image/([^;]+);base64,([^)]+))\)", markdown
    ):
        _, data_uri, image_format, base64_data = image_match.groups()

        if image_format.lower() not in IMAGE_EXTENSIONS:
            markdown = markdown.replace(image_match.group(0), "")
            warning(f"Unsupported image format: {image_format}, image will be ignored")
            continue

        image_data = base64.b64decode(base64_data)
        image_path = image_dir / (uuid.uuid4().hex[:4] + "." + image_format)

        with open(image_path, "wb") as f:
            f.write(image_data)

        # Replace data URI with relative path
        markdown = markdown.replace(data_uri, str(image_path))

    return markdown


if __name__ == "__main__":
    assert len(sys.argv) == 2, "Usage: python any2markdown.py <workspace>"
    work_dir = Path(sys.argv[1])
    assert work_dir.exists(), f"Workspace {work_dir} does not exist."
    os.chdir(work_dir)
    set_logger(
        f"any2markdown-{work_dir.stem}", work_dir / ".history" / "any2markdown.log"
    )

    mcp.run(show_banner=False)



import json
import traceback
import uuid
from collections.abc import AsyncGenerator
from pathlib import Path
from typing import Literal

from deeppresenter.agents.design import Design
from deeppresenter.agents.env import AgentEnv
from deeppresenter.agents.pptagent import PPTAgent
from deeppresenter.agents.research import Research
from deeppresenter.utils.config import DeepPresenterConfig
from deeppresenter.utils.constants import WORKSPACE_BASE
from deeppresenter.utils.log import debug, error, set_logger, timer, warning
from deeppresenter.utils.typings import ChatMessage, ConvertType, InputRequest, Role
from deeppresenter.utils.webview import PlaywrightConverter, convert_html_to_pptx


class AgentLoop:
    def __init__(
        self,
        config: DeepPresenterConfig,
        session_id: str | None = None,
        workspace: Path = None,
        language: Literal["zh", "en"] = "en",
    ):
        self.config = config
        self.language = language
        if session_id is None:
            session_id = str(uuid.uuid4())[:8]
        self.workspace = workspace or WORKSPACE_BASE / session_id
        self.intermediate_output = {}
        self.agent = None
        set_logger(
            f"deeppresenter-loop-{self.workspace.stem}",
            self.workspace / ".history" / "deeppresenter-loop.log",
        )
        debug(f"Initialized AgentLoop with workspace={self.workspace}")
        debug(f"Config: {self.config.model_dump_json(indent=2)}")

    @timer("DeepPresenter Loop")
    async def run(
        self,
        request: InputRequest,
        check_llms: bool = False,
        soft_parsing: bool = True,
    ) -> AsyncGenerator[str | ChatMessage, None]:
        """Main loop for DeepPresenter generation process.
        Arguments:
            request: InputRequest object containing task details.
            check_llms: Whether to check LLM availability before running.
            soft_parsing: Whether to use soft parsing on html2pptx.
        Yields:
            ChatMessage or str: Messages or final output path.
        """
        if not self.config.design_agent.is_multimodal and self.config.heavy_reflect:
            debug(
                "Reflective design requires a multimodal LLM in the design agent, reflection will only enable on textual state."
            )
        if check_llms:
            await self.config.validate_llms()
        request.copy_to_workspace(self.workspace)
        with open(self.workspace / ".input_request.json", "w") as f:
            json.dump(request.model_dump(), f, ensure_ascii=False, indent=2)
        async with AgentEnv(self.workspace, self.config) as agent_env:
            self.agent_env = agent_env
            hello_message = f"DeepPresenter running in {self.workspace}, with {len(request.attachments)} attachments, prompt={request.instruction}"
            if self.config.offline_mode:
                hello_message += " [Offline Mode]"
            debug(hello_message)
            yield ChatMessage(role=Role.SYSTEM, content=hello_message)
            self.research_agent = Research(
                self.config,
                agent_env,
                self.workspace,
                self.language,
            )
            self.agent = self.research_agent
            try:
                async for msg in self.research_agent.loop(request):
                    if isinstance(msg, str):
                        md_file = Path(msg)
                        if not md_file.is_absolute():
                            md_file = self.workspace / md_file
                        self.intermediate_output["manuscript"] = md_file
                        msg = str(md_file)
                        break
                    yield msg
            except Exception as e:
                error_message = (
                    f"Research agent failed with error: {e}\n{traceback.format_exc()}"
                )
                error(error_message)
                raise e
            finally:
                self.research_agent.save_history()
                self.save_results()
            if request.convert_type == ConvertType.PPTAGENT:
                self.pptagent = PPTAgent(
                    self.config,
                    agent_env,
                    self.workspace,
                    self.language,
                )
                self.agent = self.pptagent
                try:
                    async for msg in self.pptagent.loop(request, md_file):
                        if isinstance(msg, str):
                            pptx_file = Path(msg)
                            if not pptx_file.is_absolute():
                                pptx_file = self.workspace / pptx_file
                            self.intermediate_output["pptx"] = pptx_file
                            self.intermediate_output["final"] = pptx_file
                            msg = str(pptx_file)
                            break
                        yield msg
                except Exception as e:
                    error_message = (
                        f"PPTAgent failed with error: {e}\n{traceback.format_exc()}"
                    )
                    error(error_message)
                    raise e
                finally:
                    self.pptagent.save_history()
                    self.save_results()
            else:
                self.designagent = Design(
                    self.config,
                    agent_env,
                    self.workspace,
                    self.language,
                )
                self.agent = self.designagent
                try:
                    async for msg in self.designagent.loop(request, md_file):
                        if isinstance(msg, str):
                            slide_html_dir = Path(msg)
                            if not slide_html_dir.is_absolute():
                                slide_html_dir = self.workspace / slide_html_dir
                            self.intermediate_output["slide_html_dir"] = slide_html_dir
                            break
                        yield msg
                except Exception as e:
                    error_message = (
                        f"Design agent failed with error: {e}\n{traceback.format_exc()}"
                    )
                    error(error_message)
                    raise e
                finally:
                    self.designagent.save_history()
                    self.save_results()
                pptx_path = self.workspace / f"{md_file.stem}.pptx"
                try:
                    # ? this feature is in experimental stage
                    await convert_html_to_pptx(
                        slide_html_dir,
                        pptx_path,
                        aspect_ratio=request.powerpoint_type,
                        soft_parsing=soft_parsing,
                    )
                except Exception as e:
                    warning(
                        f"html2pptx conversion failed, falling back to pdf conversion\n{e}"
                    )
                    pptx_path = pptx_path.with_suffix(".pdf")
                    (self.workspace / ".html2pptx-error.txt").write_text(
                        str(e) + "\n" + traceback.format_exc()
                    )
                finally:
                    async with PlaywrightConverter() as pc:
                        await pc.convert_to_pdf(
                            list(slide_html_dir.glob("*.html")),
                            pptx_path.with_suffix(".pdf"),
                            aspect_ratio=request.powerpoint_type,
                        )

                self.intermediate_output["final"] = str(pptx_path)
                msg = pptx_path
            self.save_results()
            debug(f"DeepPresenter finished, final output at: {msg}")
            yield msg

    def save_results(self):
        with open(self.workspace / "intermediate_output.json", "w") as f:
            json.dump(
                {k: str(v) for k, v in self.intermediate_output.items()},
                f,
                ensure_ascii=False,
                indent=2,
            )



"""Convert a PDF to a single vertically-stitched image."""

import argparse

from pdf2image import convert_from_path
from PIL import Image


def pdf_to_long_image(
    pdf_path: str, image_path: str, dpi: int = 200, first_k: int | None = None
) -> None:
    pages = convert_from_path(pdf_path, dpi=dpi, last_page=first_k)
    width = max(p.width for p in pages)
    height = sum(p.height for p in pages)

    result = Image.new("RGB", (width, height))
    y = 0
    for page in pages:
        result.paste(page, (0, y))
        y += page.height

    result.save(image_path)
    print(f"Saved {len(pages)} pages -> {image_path} ({width}x{height})")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("pdf_path")
    parser.add_argument("image_path")
    parser.add_argument("--dpi", type=int, default=200)
    parser.add_argument("--first-k", type=int, default=None)
    args = parser.parse_args()
    pdf_to_long_image(args.pdf_path, args.image_path, args.dpi, args.first_k)



import os
import re
import sys
from pathlib import Path
from typing import Any, Literal

import arxiv
from fastmcp import FastMCP
from semanticscholar import AsyncSemanticScholar

from deeppresenter.utils.log import set_logger

mcp = FastMCP(name="Research")
PAGE_SIZE = int(os.getenv("ARXIV_PAGE_SIZE", "5"))
client = arxiv.Client(page_size=PAGE_SIZE)
sch = AsyncSemanticScholar()


@mcp.tool()
def search_papers(
    query: str,
    max_results: int | None = 5,
) -> dict[str, Any]:
    """
    Search for academic papers using arXiv query syntax.
    Best Practice: Start with broad queries and progressively add constraints if too many results are returned.

    QUERY PARAMETERS (field prefixes):
    - ti: Title field, e.g., ti:"attention mechanism"
    - au: Author name, e.g., au:"Hinton"
    - abs: Abstract content, e.g., abs:"transformer"
    - cat: Subject category, e.g., cat:cs.AI

    Args:
        query: Search query using arXiv syntax (keywords, field searches, boolean operators)
        max_results: Maximum number of results to return (default: 5)

    Returns:
        List of papers, each containing: title, authors, abstract, published date and pdf_url
    """
    # Map sort_by string to arxiv SortCriterion
    search = arxiv.Search(
        query=query,
        max_results=PAGE_SIZE,
    )

    papers = [
        {
            "title": paper.title,
            "authors": [author.name for author in paper.authors],
            "summary": paper.summary,
            "published": paper.published.strftime("%Y-%m-%d"),
            "pdf_url": paper.pdf_url,
        }
        for paper in client.results(search)
    ]
    if max_results is not None:
        papers = papers[:max_results]
    return {"total_results": len(papers), "papers": papers}


@mcp.tool()
async def get_paper_authors(arxiv_id: str) -> dict:
    """
    Get the authors of a paper by arxiv id.
    Args:
        arxiv_id: The id of the paper, must be in the format of ARXIV:2501.03936
    Returns:
        List of authors, containing their authorId and details.
    """
    assert re.fullmatch(r"ARXIV:\d{4}\.\d{4,5}(v\d+)?", arxiv_id), (
        "Invalid arxiv_id format. It should be like ARXIV:2501.03936"
    )
    fields = ["name", "citationCount", "affiliations"]
    authors = await sch.get_paper_authors(arxiv_id, fields=fields)
    return {"authors": [author._data for author in authors]}


@mcp.tool()
async def get_scholar_details(
    author_id: str,
    paper_start_index: int = 0,
    sort_by: Literal["citationCount", "year"] | None = None,
) -> dict[str, Any]:
    """
    Get the details of a scholar by author ID.

    Args:
        author_id: The Semantic Scholar author ID of the scholar
        paper_start_index: Starting index for paper pagination (default: 0)
        sort_by: Sort papers by "citationCount" or "year" (default: None - no sorting)

    Returns:
        Dictionary containing scholar details including:
        - Basic info: name, hIndex, homepage, paperCount
        - papers: List of papers with title, year, citationCount, publicationVenue, arxivId
    """
    fields = [
        "hIndex",
        "homepage",
        "paperCount",
        "papers",
        "papers.citationCount",
        "papers.externalIds",
        "papers.publicationDate",
        "papers.publicationVenue",
        "papers.title",
        "papers.year",
        "name",
    ]
    author = await sch.get_author(author_id, fields=fields)
    author, papers = author._data, author._data.pop("papers")
    processed_papers = []
    for p in papers:
        p["arxivId"] = p.pop("externalIds", {}).get("ArXiv", "")
        p.pop("paperId", "")
        processed_papers.append({k: v for k, v in p.items() if v is not None})
    if sort_by == "citationCount":
        processed_papers.sort(key=lambda x: x.get("citationCount", 0), reverse=True)
    elif sort_by == "year":
        processed_papers.sort(key=lambda x: x.get("year", 0), reverse=True)
    author["papers"] = processed_papers[
        paper_start_index : paper_start_index + PAGE_SIZE
    ]
    return author


if __name__ == "__main__":
    assert len(sys.argv) == 2, "Usage: python research.py <workspace>"
    work_dir = Path(sys.argv[1])
    assert work_dir.exists(), f"Workspace {work_dir} does not exist."
    os.chdir(work_dir)
    set_logger(f"research-{work_dir.stem}", work_dir / ".history" / "research.log")

    mcp.run(show_banner=False)



import asyncio
import json
import os
from glob import glob
from os.path import join

from pptagent.induct import SlideInducter
from pptagent.model_utils import ModelManager
from pptagent.multimodal import ImageLabler
from pptagent.presentation import Presentation
from pptagent.utils import Config, ppt_to_images

pr_folders = glob("data/*/pptx/*")

model_manager = ModelManager()


async def test_induct(pr_folder: str, sem: asyncio.Semaphore):
    config = Config(pr_folder)
    prs = Presentation.from_file(join(pr_folder, "original.pptx"), config)
    prs.save(join(pr_folder, "source.pptx"))
    async with sem:
        if len(glob(join(pr_folder, "slide_images", "*"))) == 0:
            await ppt_to_images(
                join(pr_folder, "source.pptx"),
                join(pr_folder, "slide_images"),
            )
        if len(glob(join(pr_folder, "template_images", "*"))) == 0:
            prs.save(join(pr_folder, "template.pptx"), layout_only=True)
            await ppt_to_images(
                join(pr_folder, "template.pptx"), join(pr_folder, "template_images")
            )
        prs = Presentation.from_file(join(pr_folder, "source.pptx"), config)
        labler = ImageLabler(prs, config)
        try:
            labler.apply_stats(
                json.load(open(join(pr_folder, "image_stats.json"), encoding="utf-8"))
            )
        except:  # noqa: E722
            caption = await labler.caption_images_async(model_manager.vision_model)
            with open(join(pr_folder, "image_stats.json"), "w", encoding="utf-8") as f:
                json.dump(caption, f, indent=4, ensure_ascii=False)
            labler.apply_stats(
                json.load(open(join(pr_folder, "image_stats.json"), encoding="utf-8"))
            )
        if not os.path.exists(join(pr_folder, "slide_induction.json")):
            inducter = SlideInducter(
                prs,
                join(pr_folder, "slide_images"),
                join(pr_folder, "template_images"),
                config,
                model_manager.image_model,
                model_manager.language_model,
                model_manager.vision_model,
            )
            reference = await inducter.content_induct(await inducter.layout_induct())
            with open(
                join(pr_folder, "slide_induction.json"), "w", encoding="utf-8"
            ) as f:
                json.dump(reference, f, indent=4, ensure_ascii=False)
        print(pr_folder, "done")


async def main():
    sem = asyncio.Semaphore(16)
    async with asyncio.TaskGroup() as tg:
        for pr_folder in pr_folders:
            tg.create_task(test_induct(pr_folder, sem))


if __name__ == "__main__":
    asyncio.run(main())



import functools
import inspect
import logging
import time
import traceback
from collections.abc import Callable, Coroutine
from contextvars import ContextVar
from pathlib import Path
from typing import Any, ParamSpec, TypeVar, overload

import colorlog
from openai import (
    APIConnectionError,
    APIError,
    APIResponseValidationError,
    APIStatusError,
    APITimeoutError,
    AuthenticationError,
    BadRequestError,
    ConflictError,
    ContentFilterFinishReasonError,
    InternalServerError,
    InvalidWebhookSignatureError,
    LengthFinishReasonError,
    NotFoundError,
    OpenAIError,
    PermissionDeniedError,
    RateLimitError,
    UnprocessableEntityError,
)
from pydantic import ValidationError

from deeppresenter.utils.constants import LOGGING_LEVEL

_context_logger: ContextVar[logging.Logger | None] = ContextVar(
    "_context_logger", default=None
)
P = ParamSpec("P")
R = TypeVar("R")


def create_logger(
    name: str = __name__, log_file: str | Path | None = None
) -> logging.Logger:
    """Create a new logger"""
    assert name == "default logger" or name not in logging.Logger.manager.loggerDict, (
        f"Logger '{name}' already exists."
    )
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    # Clear existing handlers to avoid duplicates
    logger.handlers.clear()

    console_handler = logging.StreamHandler()
    console_handler.setLevel(LOGGING_LEVEL)
    formatter = colorlog.ColoredFormatter(
        "%(log_color)s%(levelname)-4s%(reset)s %(asctime)s [%(name)s] %(blue)s%(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        reset=True,
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red,bg_white",
        },
    )
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    if log_file is not None:
        path = Path(log_file)
        path.parent.mkdir(parents=True, exist_ok=True)

        file_handler = logging.FileHandler(path, encoding="utf-8")
        file_handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(levelname)-4s %(asctime)s [%(name)s] %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger


def set_logger(name: str = __name__, log_file: str | Path | None = None):
    """Set a new logger for the current async context"""
    logger = _context_logger.get()
    assert logger is None or logger.name == "default logger", (
        "Context logger is already set."
    )
    logger = create_logger(name, log_file)
    logger.debug("Setting new context logger with loglevel=%s", LOGGING_LEVEL)
    _context_logger.set(logger)
    return logger


def get_logger() -> logging.Logger:
    """Get the logger for the current context"""

    ctx_logger = _context_logger.get()
    if ctx_logger is None:
        ctx_logger = create_logger("default logger")
        _context_logger.set(ctx_logger)
    return ctx_logger


def debug(msg, *args, **kwargs):
    get_logger().debug(msg, *args, **kwargs)


def info(msg, *args, **kwargs):
    get_logger().info(msg, *args, **kwargs)


def warning(msg, *args, **kwargs):
    get_logger().warning(msg, *args, **kwargs)


def error(msg, *args, **kwargs):
    get_logger().error(msg, *args, **kwargs)


def critical(msg, *args, **kwargs):
    get_logger().critical(msg, *args, **kwargs)


def exception(msg, *args, **kwargs):
    get_logger().exception(msg, *args, **kwargs)


class timer:
    """Timer context manager and decorator"""

    def __init__(self, name: str = None):
        self.name = name
        self.start_time = None

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        elapsed = time.time() - self.start_time
        if elapsed > 1:
            debug(f"{self.name} took {elapsed:.2f} seconds")

    @overload
    def __call__(
        self, func: Callable[P, Coroutine[Any, Any, R]]
    ) -> Callable[P, Coroutine[Any, Any, R]]: ...

    @overload
    def __call__(self, func: Callable[P, R]) -> Callable[P, R]: ...

    def __call__(self, func):
        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def async_wrapper(*args, **kwargs):
                start = time.time()
                try:
                    return await func(*args, **kwargs)
                finally:
                    elapsed = time.time() - start
                    if elapsed > 1:
                        debug(
                            f"{self.name or func.__name__} took {elapsed:.2f} seconds"
                        )

            return async_wrapper
        else:

            @functools.wraps(func)
            def sync_wrapper(*args, **kwargs):
                start = time.time()
                try:
                    return func(*args, **kwargs)
                finally:
                    elapsed = time.time() - start
                    if elapsed > 1:
                        debug(
                            f"{self.name or func.__name__} took {elapsed:.2f} seconds"
                        )

            return sync_wrapper


def logging_openai_exceptions(identifider: str | Any, exc: Exception):
    """Log OpenAI exceptions"""
    if isinstance(exc, RateLimitError):
        msg = f"RateLimitError (HTTP 429): {str(exc)}"
    elif isinstance(exc, APITimeoutError):
        msg = f"APITimeoutError: {str(exc)}"
    elif isinstance(exc, APIConnectionError):
        msg = f"APIConnectionError: {str(exc)}"
    elif isinstance(exc, AuthenticationError):
        msg = f"AuthenticationError (HTTP 401): {str(exc)}"
    elif isinstance(exc, PermissionDeniedError):
        msg = f"PermissionDeniedError (HTTP 403): {str(exc)}"
    elif isinstance(exc, NotFoundError):
        msg = f"NotFoundError (HTTP 404): {str(exc)}"
    elif isinstance(exc, ConflictError):
        msg = f"ConflictError (HTTP 409): {str(exc)}"
    elif isinstance(exc, BadRequestError):
        msg = f"BadRequestError (HTTP 400): {str(exc)}"
    elif isinstance(exc, UnprocessableEntityError):
        msg = f"UnprocessableEntityError (HTTP 422): {str(exc)}"
    elif isinstance(exc, InternalServerError):
        msg = f"InternalServerError (HTTP 500): {str(exc)}"
    elif isinstance(exc, APIStatusError):
        code = getattr(exc, "status_code", "unknown")
        msg = f"APIStatusError (HTTP {code}): {str(exc)}"
    elif isinstance(exc, APIError):
        msg = f"APIError: {str(exc)}"
    elif isinstance(exc, APIResponseValidationError):
        msg = f"APIResponseValidationError: {str(exc)}"
    elif isinstance(exc, InvalidWebhookSignatureError):
        msg = f"InvalidWebhookSignatureError: {str(exc)}"
    elif isinstance(exc, ContentFilterFinishReasonError):
        msg = f"ContentFilterFinishReasonError: {str(exc)}"
    elif isinstance(exc, LengthFinishReasonError):
        msg = f"LengthFinishReasonError: {str(exc)}"
    elif isinstance(exc, OpenAIError):
        msg = f"OpenAIError: {str(exc)}"
    elif isinstance(exc, ValidationError):
        msg = f"Pydantic ValidationError: {str(exc)}"
    elif hasattr(exc, "http_status"):
        msg = f"OpenAI API Error {exc.http_status}: {str(exc)}"
    else:
        msg = f"Exception: {str(exc)}\n{traceback.format_exc()}"

    warning(f"{identifider} encountered {msg}")
    return msg
